# -*- coding: utf-8 -*-
"""
This module contains the Metaheuristic class.

Created on Thu Sep 26 16:56:01 2019

@author: Jorge Mario Cruz-Duarte (jcrvz.github.io), e-mail: jorge.cruz@tec.mx
"""

import numpy as np
from . import operators as Operators
from .population import Population

__all__ = ['Metaheuristic', 'Population', 'Operators']
__operators__ = Operators.__all__
__selectors__ = ['greedy', 'probabilistic', 'metropolis', 'all', 'none']


class Metaheuristic:
    """
        This is the Metaheuristic class, each object corresponds to a metaheuristic implemented with a sequence of
        search operators from op, and it is based on a population from Population.
    """
    def __init__(self, problem, search_operators=None, num_agents: int = 30, num_iterations: int = 100,
                 initial_scheme: str = 'random', verbose: bool = False):
        """
        Create a population-based metaheuristic by employing different simple search operators.

        :param dict problem:
            This is a dictionary containing the 'function' that maps a 1-by-D array of real values to a real value,
            'is_constrained' flag that indicates the solution is inside the search space, and the 'boundaries' (a tuple
            with two lists of size D). These two lists correspond to the lower and upper limits of domain, such as:
            ``boundaries = (lower_boundaries, upper_boundaries)``

            **Note:** Dimensions (D) of search domain are read from these boundaries. The problem can be obtained from
            the ``benchmark_func`` module.
        :param list search_operators:
            A list of available search operators. These operators must correspond to those available in the
            ``operators`` module. This parameter is mandatory for mataheuristic implementations, for using parts of this
            class, these can be provided as a list of ``operators``.
        :param int num_agents: Optional.
            Number of agents or population size. The default is 30.
        :param int num_iterations: Optional.
            Number of iterations or generations that the metaheuristic is going to perform. The default is 100.

        :return: None.
        """
        # Read the problem function
        self.finalisation_conditions = None
        self._problem_function = problem['function']

        # Create population
        self.pop = Population(problem['boundaries'], num_agents, problem['is_constrained'])

        # Check and read the search_operators
        if search_operators:
            if not isinstance(search_operators, list):
                search_operators = [search_operators]
            self.perturbators, self.selectors = Operators.process_operators(search_operators)

        # Define the maximum number of iterations
        self.num_iterations = num_iterations

        # Read the number of dimensions
        self.num_dimensions = self.pop.num_dimensions

        # Read the number of agents
        self.num_agents = num_agents

        # Initialise historical variables
        self.historical = dict()

        # Set additional variables
        self.verbose = verbose

        # Set the initial scheme
        self.initial_scheme = initial_scheme

    def apply_initialiser(self):
        # Set initial iteration
        self.pop.iteration = 0

        # Initialise the population
        self.pop.initialise_positions(self.initial_scheme)  # Default: random

        # Evaluate fitness values
        self.pop.evaluate_fitness(self._problem_function)

        # Update population, particular, and global
        self.pop.update_positions('population', 'all')  # Default
        self.pop.update_positions('particular', 'all')
        self.pop.update_positions('global', 'greedy')

    def apply_search_operator(self, perturbator, selector):
        # Split operator
        operator_name, operator_params = perturbator.split('(')

        # Apply an operator
        exec('Operators.' + operator_name + '(self.pop,' + operator_params)

        # Evaluate fitness values
        self.pop.evaluate_fitness(self._problem_function)

        # Update population
        if selector in __selectors__:
            self.pop.update_positions('population', selector)
        else:
            self.pop.update_positions()

        # Update global position
        self.pop.update_positions('global', 'greedy')

    def run(self):
        """
        Run the metaheuristic for solving the defined problem.
        :return: None.
        """
        if (not self.perturbators) or (not self.selectors):
            raise Operators.OperatorsError("There are not perturbator or selector!")

        # Apply initialiser / Random Sampling
        self.apply_initialiser()

        # Initialise and update historical variables
        self.reset_historicals()
        self.update_historicals()

        # Report which operators are going to use
        self._verbose('\nSearch operators to employ:')
        for perturbator, selector in zip(self.perturbators, self.selectors):
            self._verbose("{} with {}".format(perturbator, selector))
        self._verbose("{}".format('-' * 50))

        # Start optimisaton procedure
        while not self.finaliser():
            # Update the current iteration
            self.pop.iteration += 1

            # Implement the sequence of operators and selectors
            for perturbator, selector in zip(self.perturbators, self.selectors):

                # Apply the corresponding search operator
                self.apply_search_operator(perturbator, selector)

                # Update historical variables
                self.update_historicals()

            # Verbose (if so) some information
            self._verbose('{}\npop. radius: {}'.format(self.pop.iteration, self.historical['radius'][-1]))
            self._verbose(self.pop.get_state())

    def set_finalisation_conditions(self, conditions):
        # TODO: Check that it works for budgets <=
        if not isinstance(conditions, list):
            conditions = list(conditions)

        self.finalisation_conditions = conditions

    def finaliser(self):
        criteria = self.pop.iteration >= self.num_iterations
        if self.finalisation_conditions is not None:
            for condition in self.finalisation_conditions:
                criteria |= condition()

        return criteria

    def get_solution(self):
        """
        Deliver the last position and fitness value obtained after ``run`` the metaheuristic procedure.
        :returns: ndarray, float
        """
        return self.historical['position'][-1], self.historical['fitness'][-1]

    def reset_historicals(self):
        """
        Reset the ``historical`` variables.
        :return: None.
        """
        self.historical = dict(fitness=list(), position=list(), centroid=list(), radius=list())

    def update_historicals(self):
        """
        Update the ``historical`` variables.
        :return: None.
        """
        # Update historical variables
        self.historical['fitness'].append(np.copy(self.pop.global_best_fitness))
        self.historical['position'].append(np.copy(self.pop.global_best_position))

        # Update population centroid and radius
        current_centroid = np.array(self.pop.positions).mean(0)
        self.historical['centroid'].append(np.copy(current_centroid))
        self.historical['radius'].append(np.max(np.linalg.norm(self.pop.positions - np.tile(
            current_centroid, (self.num_agents, 1)), 2, 1)))

    def _verbose(self, text_to_print):
        """
        Print each step performed during the solution procedure. It only works if ``verbose`` flag is True.
        :param str text_to_print:
            Explanation about what the metaheuristic is doing.
        :return: None.
        """
        if self.verbose:
            print(text_to_print)



__all__ = ['benchmark_func', 'experiment', 'hyperheuristic', 'metaheuristic', 'operators', 'population', 'tools',
           'visualisation']
# Under revision: 'characterisation', 'visualisation'
__version__ = "1.1.4"



# -*- coding: utf-8 -*-
"""
This module contains the class Population.

Created on Tue Sep 17 14:29:43 2019

@author: Jorge Mario Cruz-Duarte (jcrvz.github.io), e-mail: jorge.cruz@tec.mx
"""

import numpy as np

__all__ = ['Population']
__selectors__ = ['all', 'greedy', 'metropolis', 'probabilistic']


class Population:
    """
    This is the Population class, each object corresponds to a population of agents within a search space.
    """

    # Internal variables
    iteration = 0
    rotation_matrix = []

    # Parameters per selection method
    metropolis_temperature = 1000.0
    metropolis_rate = 0.01
    metropolis_boltzmann = 1.0
    probability_selection = 0.5

    # Class initialisation
    # ------------------------------------------------------------------------
    def __init__(self, boundaries, num_agents=30, is_constrained=True):
        """
        Return a population of size ``num_agents`` within a problem domain defined by ``boundaries``.

        :param tuple boundaries:
            A tuple with two lists of size D corresponding to the lower and upper limits of search space, such as:
                boundaries = (lower_boundaries, upper_boundaries)
            Note: Dimensions of search domain are read from these boundaries.
        :param int num_agents: optional.
            Number of search agents or population size. The default is 30.
        :param bool is_constrained: optional.
            Avoid agents abandon the search space. The default is True.

        :returns: population object.
        """
        # Read number of variables or dimension
        if len(boundaries[0]) == len(boundaries[1]):
            self.num_dimensions = len(boundaries[0])
        else:
            raise PopulationError('Lower and upper boundaries must have the same length')

        # Read the upper and lower boundaries of search space
        self.lower_boundaries = np.array(boundaries[0]) if isinstance(boundaries[0], list) else boundaries[0]
        self.upper_boundaries = np.array(boundaries[1]) if isinstance(boundaries[1], list) else boundaries[1]
        self.span_boundaries = self.upper_boundaries - self.lower_boundaries
        self.centre_boundaries = (self.upper_boundaries + self.lower_boundaries) / 2.

        # Read number of agents in population
        assert isinstance(num_agents, int)
        self.num_agents = num_agents

        # Initialise positions and fitness values
        self.positions = np.full((self.num_agents, self.num_dimensions), np.nan)
        self.velocities = np.full((self.num_agents, self.num_dimensions), 0)
        self.fitness = np.full(self.num_agents, np.nan)

        # General fitness measurements
        self.global_best_position = np.full(self.num_dimensions, np.nan)
        self.global_best_fitness = float('inf')

        self.current_best_position = np.full(self.num_dimensions, np.nan)
        self.current_best_fitness = float('inf')
        self.current_worst_position = np.full(self.num_dimensions, np.nan)
        self.current_worst_fitness = -float('inf')

        self.particular_best_positions = np.full((self.num_agents, self.num_dimensions), np.nan)
        self.particular_best_fitness = np.full(self.num_agents, np.nan)

        self.previous_positions = np.full((self.num_agents, self.num_dimensions), np.nan)
        self.previous_velocities = np.full((self.num_agents, self.num_dimensions), np.nan)
        self.previous_fitness = np.full(self.num_agents, np.nan)

        self.backup_positions = np.full((self.num_agents, self.num_dimensions), np.nan)
        self.backup_velocities = np.full((self.num_agents, self.num_dimensions), np.nan)
        self.backup_fitness = np.full(self.num_agents, np.nan)
        self.backup_particular_best_positions = np.full((self.num_agents, self.num_dimensions), np.nan)
        self.backup_particular_best_fitness = np.full(self.num_agents, np.nan)

        self.is_constrained = is_constrained

        # TODO Add capability for dealing with topologies (neighbourhoods)
        # self.local_best_fitness = self.fitness
        # self.local_best_positions = self.positions

    # ===========
    # BASIC TOOLS
    # ===========

    def get_state(self):
        """
        Return a string containing the current state of the population, i.e.,
            str = 'x_best = ARRAY, f_best = VALUE'

        :returns: str
        """
        return ('x_best = ' + str(self.rescale_back(self.global_best_position)) +
                ', f_best = ' + str(self.global_best_fitness))

    def get_positions(self):
        """
        Return the current population positions. Positions are represented in a matrix of size:
            ``positions.shape() = (num_agents, num_dimensions)``

        **NOTE:** The position is rescaled from the normalised search space, i.e., [-1, 1]^num_dimensions.

        :returns: numpy.ndarray
        """
        return np.tile(self.centre_boundaries, (self.num_agents, 1)) + self.positions * np.tile(
            self.span_boundaries / 2., (self.num_agents, 1))

    def set_positions(self, positions):
        """
        Modify the current population positions. Positions are represented in a matrix of size:
            ``positions.shape() = (num_agents, num_dimensions)``

        Note: The position is rescaled to the original search space.

        :param numpy.ndarray positions:
            Population positions must have the size num_agents-by-num_dimensions array.

        :returns: numpy.ndarray
        """
        return 2. * (positions - np.tile(self.centre_boundaries, (self.num_agents, 1))) / np.tile(
            self.span_boundaries, (self.num_agents, 1))

    def revert_positions(self):
        """
        Revert the positions to the data in backup variables.
        """
        self.fitness = np.copy(self.backup_fitness)
        self.positions = np.copy(self.backup_positions)
        self.velocities = np.copy(self.backup_velocities)
        self.particular_best_fitness = np.copy(self.backup_particular_best_fitness)
        self.particular_best_positions = np.copy(self.backup_particular_best_positions)
        self.update_positions('global', 'greedy')

    def update_positions(self, level='population', selector='all'):
        """
        Update the population positions according to the level and selection scheme.

        **NOTE:** When an operator is applied (from the operators' module), it automatically replaces new positions, so
        the logic of selectors is contrary as they commonly do.

        :param str level: optional
            Update level, it can be 'population' for the entire population, 'particular' for each agent (an its
            historical performance), and 'global' for the current solution. The default is 'population'.
        :param str selector: optional
            Selection method. The selectors available are: 'greedy', 'probabilistic', 'metropolis', 'all', and 'none'.
            The default is 'all'.

        :returns: None.
        """
        # Update population positions, velocities and fitness
        if level == 'population':
            # backup the previous position to prevent losses
            self.backup_fitness = np.copy(self.previous_fitness)
            self.backup_positions = np.copy(self.previous_positions)
            self.backup_velocities = np.copy(self.previous_velocities)
            self.backup_particular_best_fitness = np.copy(self.particular_best_fitness)
            self.backup_particular_best_positions = np.copy(self.particular_best_positions)

            for agent in range(self.num_agents):
                if self._selection(self.fitness[agent], self.previous_fitness[agent], selector):
                    # if new positions are improved, then update the previous register
                    self.previous_fitness[agent] = np.copy(self.fitness[agent])
                    self.previous_positions[agent, :] = np.copy(self.positions[agent, :])
                    self.previous_velocities[agent, :] = np.copy(self.velocities[agent, :])

                else:
                    # ... otherwise,return to previous values
                    self.fitness[agent] = np.copy(self.previous_fitness[agent])
                    self.positions[agent, :] = np.copy(self.previous_positions[agent, :])
                    self.velocities[agent, :] = np.copy(self.previous_velocities[agent, :])

            # Update the current best and worst positions (forced to greedy)
            self.current_best_position = np.copy(self.positions[self.fitness.argmin(), :])
            self.current_best_fitness = np.min(self.fitness)

            self.current_worst_position = np.copy(self.positions[self.fitness.argmax(), :])
            self.current_worst_fitness = np.min(self.fitness)

        # Update particular positions, velocities and fitness
        elif level == 'particular':
            for agent in range(self.num_agents):
                if self._selection(self.fitness[agent], self.particular_best_fitness[agent], selector):
                    self.particular_best_fitness[agent] = np.copy(self.fitness[agent])
                    self.particular_best_positions[agent, :] = np.copy(self.positions[agent, :])

        # Update global positions, velocities and fitness
        elif level == 'global':
            # Perform particular updating (recursive)
            self.update_positions('particular', selector)

            # Read current global best agent
            candidate_position = np.copy(self.particular_best_positions[self.particular_best_fitness.argmin(), :])
            candidate_fitness = np.min(self.particular_best_fitness)
            if self._selection(candidate_fitness, self.global_best_fitness, selector) or np.isinf(candidate_fitness):
                self.global_best_position = np.copy(candidate_position)
                self.global_best_fitness = np.copy(candidate_fitness)

        # Raise an error
        else:
            raise PopulationError('Invalid update level')

    def evaluate_fitness(self, problem_function):
        """
        Evaluate the population positions in the problem function.

        :param function problem_function:
            A function that maps a 1-by-D array of real values to a real value.

        :returns: None.
        """
        # Read problem, it must be a callable function
        assert callable(problem_function)

        # Check simple constraints before evaluate
        if self.is_constrained:
            self._check_simple_constraints()

        # Evaluate each agent in this function
        for agent in range(self.num_agents):
            self.fitness[agent] = problem_function(self.rescale_back(self.positions[agent, :]))

    # ==============
    # INITIALISATORS
    # ==============
    # TODO Add more initialisation operators like grid, boundary, etc.

    def initialise_positions(self, scheme='random'):
        """
        Initialise population by an initialisation scheme.

        :param str scheme: optional
            Initialisation scheme. It is only 'random' and 'vertex' initialisation in the current version. We are
            working on implementing initialisation methods. The 'random' consists of using a random uniform distribution
            in [-1,1]. Otherwise, 'vertex' uses the vertices of nested hyper-cubes to allocate the agents. The default
            is 'random'.

        :returns: None.
        """
        if scheme == 'vertex':
            self.positions = self._grid_matrix(self.num_dimensions, self.num_agents)
        else:
            self.positions = np.random.uniform(-1, 1, (self.num_agents, self.num_dimensions))

    # ================
    # INTERNAL METHODS
    # ================
    # Avoid using them outside

    @staticmethod
    def _grid_matrix(num_dimensions, num_agents):

        total_vertices = 2 ** num_dimensions

        basic_matrix = 2 * np.array(
            [[int(x) for x in list(format(k, '0{}b'.format(num_dimensions)))] for k in range(total_vertices)]) - 1

        output_matrix = np.copy(basic_matrix)

        if num_agents > total_vertices:
            num_matrices = int(np.ceil((num_agents - total_vertices) / total_vertices)) + 1
            for k in range(1, num_matrices):
                k_matrix = (1 - k / num_matrices) * basic_matrix
                output_matrix = np.concatenate((output_matrix, k_matrix), axis=0)

        output_matrix = output_matrix[:num_agents, :]

        return output_matrix

    def _check_simple_constraints(self):
        """
        Check simple constraints for all the dimensions like:
            -1 <= position <= 1, for all i in 1, 2, ..., num_dimensions
        When an agent position is outside the search space, it is reallocated at the closest boundary and its velocity
        is set zero (if so).

        **NOTE:** This check is performed only if Population.is_constrained = True.

        :returns: None.
        """
        # Check if there are nans values
        if np.any(np.isnan(self.positions)):
            np.nan_to_num(self.positions, copy=False, nan=1.0, posinf=1.0, neginf=-1.0)

        # Check if agents are beyond lower boundaries
        low_check = np.less(self.positions, -1.0)
        if np.any(low_check):
            # Fix them
            self.positions[low_check] = -1.0
            self.velocities[low_check] = 0.0

        # Check if agents are beyond upper boundaries
        upp_check = np.greater(self.positions, 1.0)
        if np.any(upp_check):
            # Fix them
            self.positions[upp_check] = 1.0
            self.velocities[upp_check] = 0.0

    def rescale_back(self, position):
        """
        Rescale an agent position from [-1.0, 1.0] to the original search space boundaries per dimension.

        :param numpy.ndarray position:
            A position given by an array of 1-by-D with elements between [-1, 1].

        :returns: ndarray
        """
        return self.centre_boundaries + position * (self.span_boundaries / 2)

    def _selection(self, new, old, selector='greedy'):
        """
        Answer the question: 'should this new position be accepted?' To do so, a selection procedure is applied.

        :param numpy.ndarray new:
            A new position given by an array of 1-by-num_dimensions with elements between [-1, 1].
        :param numpy.ndarray old:
            An old position given by an array of 1-by-num_dimensions with elements between [-1, 1].
        :param str selector: optional
            A selection scheme used for deciding if the new position is kept. The default is 'greedy'.

        :returns: bool
        """
        # Greedy selection
        if selector == 'greedy':
            selection_condition = new <= old

        # Metropolis selection
        elif selector == 'metropolis':
            if new <= old:
                selection_condition = True
            else:
                selection_condition = bool(np.math.exp(-(new - old) / (
                        self.metropolis_boltzmann * self.metropolis_temperature *
                        ((1 - self.metropolis_rate) ** self.iteration) + 1e-23)) > np.random.rand())

        # Probabilistic selection
        elif selector == 'probabilistic':
            selection_condition = bool((new <= old) or (np.random.rand() <= self.probability_selection))

        # All selection
        elif selector == 'all':
            selection_condition = True

        # None selection
        elif selector == 'none':
            selection_condition = False
        else:
            selection_condition = None
            raise PopulationError('Invalid selector!')

        return selection_condition


class PopulationError(Exception):
    """
    Simple PopulationError to manage exceptions.
    """
    pass



# -*- coding: utf-8 -*-
"""
This module contains the main experiments performed using the current framework.

Created on Mon Sep 30 13:42:15 2019

@author: Jorge Mario Cruz-Duarte (jcrvz.github.io), e-mail: jorge.cruz@tec.mx
"""

from . import hyperheuristic as hyp
from . import operators as op
from . import benchmark_func as bf
from . import tools as tl
import multiprocessing
from os import path


# %% PREDEFINED CONFIGURATIONS
# Use configuration files instead of predefined dictionaries

# Configuration dictionary for experiments
ex_configs = [
    {'experiment_name': 'demo_test', 'experiment_type': 'default', 'heuristic_collection_file': 'default.txt',
     'weights_dataset_file': 'operators_weights.json'},  # 0 - Default
    {'experiment_name': 'brute_force', 'experiment_type': 'brute_force',
     'heuristic_collection_file': 'default.txt'},  # 1 - Brute force
    {'experiment_name': 'basic_metaheuristics', 'experiment_type': 'basic_metaheuristics',
     'heuristic_collection_file': 'basicmetaheuristics.txt'},  # 2 - Basic metaheuristics
    {'experiment_name': 'short_test1', 'experiment_type': 'default', 'heuristic_collection_file': 'default.txt',
     'weights_dataset_file': 'operators_weights.json'},  # 3 - Short collection
    {'experiment_name': 'short_test2', 'experiment_type': 'default',
     'heuristic_collection_file': 'default.txt'},  # 4 - Short collection +
    {'experiment_name': 'long_test', 'experiment_type': 'default', 'heuristic_collection_file': 'test-set-21.txt',
     'auto_collection_num_vals': 21}  # 5 - Long collection
]

# Configuration dictionary for hyper-heuristics
hh_configs = [
    {'cardinality': 3, 'num_replicas': 30},  # 0 - Default
    {'cardinality': 1, 'num_replicas': 30},  # 1 - Brute force
    {'cardinality': 1, 'num_replicas': 30},  # 2 - Basic metaheuristic
    {'cardinality': 3, 'num_replicas': 50},  # 3 - Short collection
    {'cardinality': 5, 'num_replicas': 50},  # 4 - Short collection +
    {'cardinality': 3, 'num_replicas': 50}   # 5 - Long collection
]

# Configuration dictionary for problems
pr_configs = [
    {'dimensions': [2, 5], 'functions': ['<choose_randomly>']},  # 0 - Default
    {'dimensions': [2, 5, *range(10, 50 + 1, 10)], 'functions': bf.__all__},    # 1 - Brute force
    {'dimensions': [2, 5, *range(10, 50 + 1, 10)], 'functions': bf.__all__},    # 2 - Basic metaheuristic
    {'dimensions': [2, 5, *range(10, 50 + 1, 10)], 'functions': bf.__all__},    # 3 - Short collection
    {'dimensions': [2, 5, *range(10, 50 + 1, 10)], 'functions': bf.__all__},    # 4 - Short collection +
    {'dimensions': [2, 5, *range(10, 50 + 1, 10)], 'functions': bf.__all__},    # 5 - Long collection
]


# %% EXPERIMENT CLASS

# _ex_config_demo = {'experiment_name': 'demo_test', 'experiment_type': 'default',
#                    'heuristic_collection_file': 'default.txt', 'weights_dataset_file': 'operators_weights.json'}
# _hh_config_demo = {'cardinality': 3, 'num_replicas': 30}
# _pr_config_demo = {'dimensions': [2, 5], 'functions': [bf.__all__[hyp.np.random.randint(0, len(bf.__all__))]]}

class Experiment:
    """
    Create an experiment using certain configurations.
    """

    def __init__(self, config_file=None, exp_config=None, hh_config=None, prob_config=None):
        """
        Initialise the experiment object.

        :param str config_file:
            Name of the configuration JSON file with the configuration dictionaries: exp_config, hh_config, and
            prob_config. If only the filename is provided, it is assumed that such a file is in the directory
            './exconf/'. Otherwise, the full path must be entered. The default value is None.

        :param dict exp_config:
            Configuration dictionary related to the experiment. Keys and default values are listed as follows:

            'experiment_name':              'test',         # Name of the experiment
            'experiment_type':              'default',      # Type: 'default', 'brute_force', 'basic_metaheuristics'
            'heuristic_collection_file':    'default.txt',  # Heuristic space located in /collections/
            'weights_dataset_file':         None,           # Weights or probability distribution of heuristic space
            'use_parallel':                 True,           # Run the experiment using a pool of processors
            'parallel_pool_size':           None,           # Number of processors available, None = Default
            'auto_collection_num_vals':     5               # Number of values for creating an automatic collection

            **NOTE 1:** 'experiment_type': 'default' or another name mean hyper-heuristic.
            **NOTE 2:** If the collection does not exist, and it is not a reserved one ('default.txt', 'automatic.txt',
            'basicmetaheuristics.txt', 'test_collection'), then an automatic heuristic space is generated with
            ``op.build_operators`` with 'auto_collection_num_vals' as ``num_vals`` and
            'heuristic_collection_file' as ``filename``.
            **NOTE 3:** # 'weights_dataset_file' must be determined in a pre-processing step. For the 'default'
            heuristic space, it is provided 'operators_weights.json'.

        :param dict hh_config:
            Configuration dictionary related to the hyper-heuristic procedure. Keys and default values are listed as
            follows:

            'cardinality':                      3,          # Maximum cardinality used for building metaheuristics
            'num_agents':                       30,         # Population size employed by the metaheuristic
            'num_iterations':                   100,        # Maximum number of iterations used by the metaheuristic
            'num_replicas':                     50,         # Number of replicas for each metaheuristic implemented
            'num_steps':                        100,        # * Number of steps that the hyper-heuristic performs
            'max_temperature':                  200,        # * Initial temperature for HH-Simulated Annealing
            'stagnation_percentage':            0.3,        # * Percentage of stagnation used by the hyper-heuristic
            'cooling_rate':                     0.05        # * Cooling rate for HH-Simulated Annealing

            **NOTE 4:** Keys with * correspond to those that are only used when ``exp_config['experiment_type']`` is
            neither 'brute_force' nor 'basic_metaheuristic'.

        :param dict prob_config:
            Configuration dictionary related to the problems to solve. Keys and default values are listed as follows:

            'dimensions':       [2, 5, 10, 20, 30, 40, 50], # List of dimensions for the problem domains
            'functions':        bf.__all__,                 # List of function names of the optimisation problems
            'is_constrained':   True                        # True if the problem domain is hard constrained

        :return: None.
        """
        self.exp_config, self.hh_config, self.prob_config = read_config_file(config_file, exp_config, hh_config,
                                                                             prob_config)

        # Check if the heuristic collection exists
        if not path.isfile('./collections/' + self.exp_config['heuristic_collection_file']):
            # If the name is a reserved one. These files cannot be not created automatically
            if exp_config['heuristic_collection_file'] in ['default.txt', 'automatic.txt', 'basicmetaheuristics.txt',
                                                           'test_collection', 'short_collection.txt']:
                raise ExperimentError('This collection name is reserved and cannot be created automatically!')
            else:
                op.build_operators(op.obtain_operators(
                    num_vals=exp_config['auto_collection_num_vals']),
                    file_name=exp_config['heuristic_collection_file'].split('.')[0])
                self.exp_config['weights_dataset_file'] = None

        # Check if the weights dataset not exist or required
        if self.exp_config['weights_dataset_file'] and (
                self.exp_config['experiment_type'] not in ['brute_force', 'basic_metaheuristics']):
            if path.isfile('collections/' + self.exp_config['weights_dataset_file']):
                self.weights_data = tl.read_json('collections/' + self.exp_config['weights_dataset_file'])
            else:
                raise ExperimentError('A valid weights_dataset_file must be provided in exp_config')
        else:
            self.weights_data = None

    def run(self):
        """
        Run the experiment according to the configuration variables.
        :return: None
        """
        # TODO: Create a task log for prevent interruptions
        # Create a list of problems from functions and dimensions combinations
        all_problems = create_task_list(self.prob_config['functions'], self.prob_config['dimensions'])

        # Check if the experiment will be in parallel
        if self.exp_config['use_parallel']:
            pool = multiprocessing.Pool(self.exp_config['parallel_pool_size'])
            pool.map(self._simple_run, all_problems)
        else:
            for prob_dim in all_problems:
                self._simple_run(prob_dim)

    def _simple_run(self, prob_dim):
        """
        Perform a single run, i.e., for a problem and dimension combination
        :param tuple prob_dim:
            Problem name and dimensionality ``(function_string, num_dimensions)``
        :return: None.
        """
        # Read the function name and the number of dimensions
        function_string, num_dimensions = prob_dim

        # Message to print and to store in folders
        label = '{}-{}D-{}'.format(function_string, num_dimensions, self.exp_config['experiment_name'])

        # Get and format the problem
        problem = eval('bf.{}({})'.format(function_string, num_dimensions))
        problem_to_solve = problem.get_formatted_problem(self.prob_config['is_constrained'])

        # Read the particular weights array (if so)
        weights = self.weights_data[str(num_dimensions)][problem.get_features(fmt='string', wrd='1')] \
            if self.weights_data else None

        # Call the hyper-heuristic object
        hh = hyp.Hyperheuristic(heuristic_space=self.exp_config['heuristic_collection_file'],
                                problem=problem_to_solve, parameters=self.hh_config,
                                file_label=label, weights_array=weights)

        # Run the HH according to the specified type
        save_steps = self.exp_config['save_steps']
        if self.exp_config['experiment_type'] in ['brute_force', 'bf']:
            hh.brute_force(save_steps)
        elif self.exp_config['experiment_type'] in ['basic_metaheuristics', 'bmh']:
            hh.basic_metaheuristics(save_steps)
        elif self.exp_config['experiment_type'] in ['online_learning', 'dynamic']:
            hh.solve('dynamic', save_steps)
        elif self.exp_config['experiment_type'] in ['neural_network']:
            hh.solve('neural_network', save_steps)
        else:  # 'static_run'
            hh.solve('static', save_steps)

        if self.exp_config['verbose']:
            print(label + ' done!')


class ExperimentError(Exception):
    """
    Simple ExperimentError to manage exceptions.
    """
    pass


def read_config_file(config_file=None, exp_config=None, hh_config=None, prob_config=None):
    """
    Return the experimental (`exp_config`), hyper-heuristic (`hh_config`), problem (`prob_config`) configuration
    variables from `config_file`, if it is supplied. Otherwise, use the `exp_config`, `hh_config`, and `prob_config`
    inputs. If there is no input, then assume the default values for these three configuration variables. Further
    information about these variables can be found in the Experiment class's `__doc__`.
    """

    # If there is a configuration file
    if config_file:
        # Adjustments
        directory, filename = path.split(config_file)
        if directory == '':
            directory = './exconf/'  # Default directory
        basename, extension = path.splitext(filename)
        if extension not in ['.json', '']:
            raise ExperimentError("Configuration file must be JSON")

        # Existence verification and load
        full_path = path.join(directory, basename + '.json')
        if path.isfile(full_path):
            all_configs = tl.read_json(full_path)

            # Load data from json file
            exp_config = all_configs['ex_config']
            hh_config = all_configs['hh_config']
            prob_config = all_configs['prob_config']
        else:
            raise ExperimentError(f"File {full_path} does not exist!")
    else:
        if exp_config is None:
            exp_config = dict()
        if hh_config is None:
            hh_config = dict()
        if prob_config is None:
            prob_config = dict()

    # Load the default experiment configuration and compare it with exp_cfg
    exp_config = tl.check_fields(
        {
            'experiment_name': 'test',
            'experiment_type': 'default',  # 'default' -> hh, 'brute_force', 'basic_metaheuristics'
            'heuristic_collection_file': 'default.txt',
            'weights_dataset_file': None,  # 'operators_weights.json',
            'use_parallel': True,
            'parallel_pool_size': None,  # Default
            'auto_collection_num_vals': 5,
            'save_steps': True,
            'verbose': True,
        }, exp_config)

    # Load the default hyper-heuristic configuration and compare it with hh_cfg
    hh_config = tl.check_fields(
        {
            'cardinality': 3,
            'num_agents': 30,
            'num_iterations': 100,
            'num_replicas': 50,
            'num_steps': 200,
            'max_temperature': 1,
            'min_temperature': 1e-6,
            'stagnation_percentage': 0.37,
            'temperature_scheme': 'fast',
            'acceptance_scheme': 'exponential',
            'cooling_rate': 1e-3,
            'cardinality_min': 1,
            'repeat_operators': True,
            'as_mh': True,
            'verbose': False,
            'verbose_statistics': False,
            'trial_overflow': True,
            'learnt_dataset': None,
            'allow_weight_matrix': True,
            'learning_portion': 0.37,
            'solver': 'static',
            'tabu_idx': None,
            'model_params': None,
            'limit_time': None,
        }, hh_config)

    # Load the default problem configuration and compare it with prob_config
    prob_config = tl.check_fields(
        {
            'dimensions': [2, 5, *range(10, 50 + 1, 10)],
            'functions': bf.__all__,
            'is_constrained': True,
            'features': ['Differentiable', 'Separable', 'Unimodal']
        }, prob_config)

    # Check if there is a special case of function name: <choose_randomly>
    prob_config['functions'] = [
        bf.__all__[hyp.np.random.randint(0, len(bf.__all__))] if fun in ['<choose_randomly>', '<random>'] else fun
        for fun in prob_config['functions']]

    return exp_config, hh_config, prob_config


def create_task_list(function_list, dimension_list):
    """
    Return a list of combinations (in tuple form) for problems from functions and dimensions.
    :param list function_list:
        List of functions from the `benchmark_func` module.
    :param list dimension_list:
        List of dimensions considered for each one of these functions.
    :return: list of tuples
    """
    return [(x, y) for x in function_list for y in dimension_list]


# %% Auto-run
if __name__ == '__main__':
    # Import module for calling this code from command-line
    import argparse
    import os
    from tools import preprocess_files

    DATA_FOLDER = "./data_files/raw"
    OUTPUT_FOLDER = "./data_files/exp_output"

    # Only one argument is allowed: the code
    parser = argparse.ArgumentParser(
        description="Run certain experiment, default experiment is './exconf/demo.json'")
    parser.add_argument('-b', '--batch', action='store_true', help="carry out a batch of experiments")
    parser.add_argument('exp_config', metavar='config_filename', type=str, nargs='?', default='demo',
                        help='''Name of the configuration file in './exconf/' or its full path. Only JSON files.
                                If --batch flag is given, it is assumed that the entered file contains a list of all the
                                paths of experiment files (JSON) to carry out. It would be read as plain text.''')

    args = parser.parse_args()

    if not os.path.exists(OUTPUT_FOLDER):
        os.makedirs(OUTPUT_FOLDER)

    if args.batch:
        with open(args.exp_config, encoding='utf-8') as configs:
            exp_filenames = configs.read()
        exp_filenames = [filename.strip() for filename in exp_filenames.splitlines()]
    else:
        exp_filenames = [args.exp_config]

    for exp_filename in exp_filenames:
        tail_message = f" - ({exp_filenames.index(exp_filename)+1}/{len(exp_filenames)})" + "\n" + ("=" * 50) \
            if args.batch else ""

        print(f"\nRunning {exp_filename.split('.')[0]}" + tail_message)

        # Create the experiment to runs
        exp = Experiment(config_file=exp_filename)

        # print("* Experiment configuration: \n", "-" * 30 + "\n", json.dumps(exp.prob_config, indent=2, default=str))
        # print("* Hyper-heuristic configuration: \n", "-" * 30 + "\n",json.dumps(exp.hh_config, indent=2, default=str))
        # print("* Problem configuration: \n", "-" * 30 + "\n", json.dumps(exp.prob_config, indent=2, default=str))

        # Run the experiment et voilà
        exp.run()

        # After run, it preprocesses all the raw data
        print(f"\nPreprocessing {exp_filename.split('.')[0]}" + tail_message)
        preprocess_files("data_files/raw/",
                         kind=exp.hh_config["solver"],
                         experiment=exp_filename,
                         output_name=OUTPUT_FOLDER + "/" + exp.exp_config["experiment_name"])

        # Rename the raw folder to raw-$exp_name$
        print(f"\nChanging folder name of raw results...")
        os.rename(DATA_FOLDER, DATA_FOLDER + "-" + exp.exp_config["experiment_name"])

    print("\nExperiment(s) finished!")



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 26 16:56:01 2019

@author: Jorge Mario Cruz-Duarte (jcrvz.github.io)
"""

import numpy as np
from sklearn.neighbors import KernelDensity
from sklearn.model_selection import GridSearchCV
import scipy.stats as st


class Characteriser():

    # _normal_scaling_factor = np.power(1 / (4 * np.pi), 1/10)

    def __init__(self):
        # Define the parameters
        self.bandwidth = 1
        self.kde_samples = 1000
        self.num_repetitions = 30
        self.normalised_boundaries = True
        self.num_dimensions = None

        self.num_samples = 1000
        self.sampling_method = 'levy_walk'
        self.levy_walk_initial = 'rand'
        self.levy_walk_alpha = 0.5
        self.levy_walk_beta = 1.0

        self.position_samples = None
        self.fitness_values = None

    # Initialise the sampling method and the function evaluation
    def initialise(self, function, samples=None):
        # Read boundaries and determine the span and centre
        lower_boundaries = function.max_search_range
        upper_boundaries = function.min_search_range
        span_boundaries = upper_boundaries - lower_boundaries
        centre_boundaries = (upper_boundaries + lower_boundaries) / 2.

        # Read the number of dimensions
        self.num_dimensions = len(centre_boundaries)

        # Generate the samples
        # TODO: Add more methods for generating samples
        if samples is None:
            if self.sampling_method == 'levy_walk':
                self.position_samples = self._levy_walk(self.levy_walk_initial, self.num_samples,
                                                        self.levy_walk_alpha, self.levy_walk_beta)

        # Evaluate them in the function
        self.fitness_values = self._evaluate_positions(
            function, span_boundaries, centre_boundaries, self.position_samples)

    def length_scale(self, function=None, bandwidth_mode='silverman_rule', samples=None, kde_samples=1000):
        # Samples from the estimated pde
        self.kde_samples = kde_samples

        # Initialise the sample positions and their fitness
        if ((self.position_samples is None) | (self.fitness_values is None)) & (function is not None):
            self.initialise(function, samples)

        # Determine the length scale
        indices_1 = np.random.permutation(len(self.fitness_values))
        indices_2 = np.array([*indices_1[1:], indices_1[0]])

        length_scale = (np.abs(self.fitness_values[indices_1] - self.fitness_values[indices_2]) / np.linalg.norm(
            self.position_samples[indices_1, :] - self.position_samples[indices_2, :], axis=1)).reshape(-1, 1)

        # Estimate the bandwidth
        if not isinstance(bandwidth_mode, str):
            self.bandwidth = bandwidth_mode
        else:
            if bandwidth_mode == 'exhaustive':
                order = int(np.ceil(np.log10(np.std(length_scale))))
                coarse_grid = GridSearchCV(KernelDensity(),
                                           {'bandwidth': np.logspace(order / 2 - 3, order / 2 + 3, 25)}, cv=3)
                first_approach = coarse_grid.fit(length_scale).best_estimator_.bandwidth
                fine_grid = GridSearchCV(KernelDensity(), {'bandwidth':
                    np.linspace(0.5 * first_approach, 2 * first_approach, 50)}, cv=3)
                self.bandwidth = fine_grid.fit(length_scale).best_estimator_.bandwidth
            elif bandwidth_mode == 'scott_rule':
                self.bandwidth = 1.06 * np.std(length_scale) * np.power(self.num_samples, -1/5)
            elif bandwidth_mode == 'silverman_rule':
                self.bandwidth = 0.9 * np.min([np.std(length_scale), st.iqr(length_scale)/1.34]) * \
                                 np.power(self.num_samples, -1/5)
            else:
                self.bandwidth = None

        # Estimate the distribution function
        pdf_xvalues = np.linspace(0.9 * length_scale.min(), 1.1 * length_scale.max(), self.kde_samples).reshape(-1, 1)
        pdf_fvalues = np.exp(KernelDensity(bandwidth=self.bandwidth).fit(length_scale).score_samples(pdf_xvalues))

        # Get statistics from raw length_scale values
        dst = st.describe(length_scale)

        # Determine the entropy metric
        entropy_value = (pdf_xvalues[1] - pdf_xvalues[0]) * st.entropy(pdf_fvalues, base=2)

        # Return a dictionary with all the information
        return dict(nob=dst.nobs,
                    raw=length_scale,
                    Min=dst.minmax[0],
                    Max=dst.minmax[1],
                    Avg=dst.mean,
                    Std=np.std(length_scale),
                    Skw=dst.skewness,
                    Kur=dst.kurtosis,
                    IQR=st.iqr(length_scale),
                    Med=np.median(length_scale),
                    MAD=st.median_absolute_deviation(length_scale),
                    KDE_bw=self.bandwidth,
                    PDF_fx=pdf_fvalues,
                    PDF_xs=pdf_xvalues,
                    Entropy=entropy_value)

    @staticmethod
    def _evaluate_positions(function, span_boundaries, centre_boundaries, positions):
        return np.array([function.get_function_value(centre_boundaries + position * (span_boundaries / 2.))
                         for position in positions])

    @staticmethod
    def _normalise_vector(vector):
        return vector / np.max([np.linalg.norm(vector), 1e-23])

    def _levy_walk(self, initial_position, num_steps=1000, alpha=0.5, beta=1.0):

        # Initial position and all the positions are normalised between -1 and 1
        if initial_position == 'rand':
            initial_position = np.random.uniform(-1, 1, self.num_dimensions)
        else:
            if not len(initial_position) == self.num_dimensions:
                raise CharacteriserError('Provide a proper initial position')

        # Initialise the output matrix
        positions = [initial_position]

        # Start the loop for all the steps
        while len(positions) <= num_steps + 1:
            # Get the Levy-distributed step and a point in the hyper-sphere surface
            new_position = positions[-1] + st.levy_stable.rvs(
                alpha, beta, size=self.num_dimensions) * self._normalise_vector(
                np.random.randn(self.num_dimensions))

            # Check if this position is within the domain and register it
            if (new_position > -1.0).all() & (new_position < 1.0).all():
                positions.append(new_position)

        return np.array(positions)




class CharacteriserError(Exception):
    """
    Simple CharacteriserError to manage exceptions.
    """
    pass


if __name__ == '__main__':
    import benchmark_func as bf
    import matplotlib.pyplot as plt

    # results_all = []
    #
    # for problem_str in bf.__all__:
    #     problem = eval('bf.' + problem_str + '(2)')
    #
    #     chsr = Characteriser()
    #     results_all.append(chsr.length_scale(problem, bandwidth_mode='silverman_rule')['Entropy'])
    #
    #     print('Evaluated ' + problem_str + '...')
    #
    # plt.semilogy([res + 1 for res in results_all]), plt.show()

    problem = bf.Sphere(2)
    problem.set_search_range(-5, 5)

    chsr = Characteriser()
    results = chsr.length_scale(problem, bandwidth_mode='exhaustive')
    plt.hist(results['raw'], density=True, bins=100), plt.plot(results['PDF_xs'], results['PDF_fx']), plt.show()

    print(results['Entropy'])

# def fast_univariate_bandwidth_estimate_STEPI(
#         num_points, source_points,  accuracy=1e-3):
#
#     # Normalise data to the unit interval
#     normalised_source_points = (source_points - np.min(source_points)) / np.max(source_points)
#
#     # Estimate the standard deviation of data
#     sigma = np.std(normalised_source_points)
#
#     # Density functional Phi_6 and Phi_8 via the normal scale rule
#     phi6 = -15 * np.power(sigma, -7) / (16 * np.sqrt(np.pi))
#     phi8 = -105 * np.power(sigma, -9) / (32 * np.sqrt(np.pi))
#
#     g1 = np.power(-6 / (np.sqrt(2 * np.pi) * phi6 * num_points), 1 / 7)
#     g1 = np.power(30 / (np.sqrt(2 * np.pi) * phi8 * num_points), 1 / 9)



# -*- coding: utf-8 -*-
"""
This module contains Machine Learning tools.

Created on Wed Sep  8 00:00:00 2021

@author: Jose Manuel Tapia Avitia, e-mail: josetapia@exatec.tec.mx
"""

import numpy as np
import pandas as pd
import tensorflow as tf
from os.path import exists as _check_path
from os import makedirs as _create_path
from timeit import default_timer as timer


def obtain_sample_weight(sample_fitness, fitness_to_weight='rank'):
    """
    Using decreasing functions to give more priority to samples with less fitness
    :param list sample_fitness: The fitness associated value for each sample
    :param str fitness_to_weight: Specify which function use to convert fitness to weight
    :return: An array that associates a weight to each sample
    """
    a = min(sample_fitness)
    b = max(sample_fitness)
    if fitness_to_weight == 'linear_reciprocal':
        # f: [a, b] -> (0, 1]
        weight_conversion = lambda fitness: a / fitness
    elif fitness_to_weight == 'linear_reciprocal_translated':
        # f: [a, b] -> [1, b-a+1]
        weight_conversion = lambda fitness: a * b / fitness - a + 1
    elif fitness_to_weight == 'linear_percentage':
        # f: [a, b] -> [1, 100]
        weight_conversion = lambda fitness: 100 * (b - fitness) / (b - a) + 1
    elif fitness_to_weight == 'rank':
        # f: [fitness] -> [0, n-1]
        indices = list(range(len(sample_fitness)))
        indices.sort(key = lambda idx: -sample_fitness[idx])
        indices_sorted = [0 for _ in indices]
        for i, idx in enumerate(indices):
            indices_sorted[idx] = i
        return indices_sorted
    else:
        # Default linear conversion
        # f: [a, b] -> [a, b]
        weight_conversion = lambda fitness: a + b - fitness
    
    return [weight_conversion(fitness) for fitness in sample_fitness]


class DatasetSequences():
    def __init__(self, sequences, fitnesses, num_operators=None, fitness_to_weight=None):
        'Pre-process sequences to generate training data for HHNN'
        X, y, sample_fitness = [], [], []
        for sequence, fitness in zip(sequences, fitnesses):
            if len(sequence) >0 and sequence[0] == -1:
                sequence.pop(0)
                fitness.pop(0)
            while len(sequence) > 0:
                # Per each prefix, predict the next operator
                y.append(sequence.pop())
                X.append(sequence.copy())
                sample_fitness.append(fitness.pop())            
        self._X = X
        self._y = y
        if fitness_to_weight is not None:
            self._sample_weight =\
                tf.constant(obtain_sample_weight(sample_fitness, fitness_to_weight))
        else:
            self._sample_weight = None
        self._one_hot_encoded = None
        if num_operators is not None:
            self.apply_one_hot_encoding(num_operators)
    
    def apply_one_hot_encoding(self, num_operators):
        'One-Hot encode the output of the training data'
        if self._one_hot_encoded is not None:
            if self._one_hot_encoded != num_operators:
                self._y = [np.where(y_one==1)[0][0] for y_one in self._y]
            else:
                return
        self._one_hot_encoded = num_operators
        if num_operators is not None:
            self._y = tf.one_hot(indices=self._y,
                                 depth=num_operators,
                                 dtype=tf.int64).numpy()
    
    def obtain_dataset(self):
        'Retrieve the pre-processed data'
        return self._X, self._y, self._sample_weight


def retrieve_model_info(params):
    # Check essential attributes
    essential_attributes = ['file_label', 'model_architecture', 'encoder', 'num_steps']
    if not all(attribute in params for attribute in essential_attributes):
        left_attributes = [attribute for attribute in essential_attributes if attribute not in params]
        raise Exception(f'The following attributes left while retrieving the model info: {left_attributes}')
    
    # Names
    architecture_name = params['model_architecture']
    encoder_name = params['encoder']

    # Model label
    memory_length = params.get('memory_length', params['num_steps'])
    attribute_labels = [params['file_label'], f'mem{memory_length}']
    if 'pretrained_model' in params:
        attribute_labels.append(params['pretrained_model'])
    model_label = '-'.join(attribute_labels)
    personal_label = params['file_label'].split('_')
    if personal_label[-1] == 'extended':
        personal_label.pop()
    personal_labels = '_'.join(personal_label)

    # Filenames
    model_directory = './data_files/ml_models/'
    model_filename = model_label
    if architecture_name == 'transformer':
        model_directory = model_directory +f'{model_label}_dir/'
        model_filename = 'trained_model'

    filename_dict = dict({
        'model_directory': model_directory,
        'model_label': model_label,
        'model_path': model_directory + f'{personal_labels}.h5',
        'log_path': model_directory + f'{model_filename}_log.csv',
        'log_time_path': model_directory + f'{model_filename}_log_time.csv'
    })
    
    return architecture_name, encoder_name, filename_dict


class Encoder():
    def __init__(self, params):
        self._encoder_name = params['encoder']
        self._architecture_name = params['model_architecture']
        self._memory_length = params.get('memory_length', 100)
        self._num_operators = params['num_operators']

        def one_hot_encode(sequence):
            return tf.one_hot(indices=sequence, depth=self._num_operators,
                              dtype=tf.int64).numpy()
        def compose(f, g):
            return lambda x: f(g(x))
        
        # Choice identity encoder
        if self._architecture_name in ['transformer', 'transformer_orig', 'LSTM_Ragged']:
            # Keep original values but element -1
            self.__identity_encoder = self.__clean_sequence
        else:
            # Keep original values but fix the length
            self.__identity_encoder = compose(self.__fix_sequence_length, self.__clean_sequence)

        # Get encoder module
        if self._encoder_name in ['identity', 'default']:
            encoder = self.__identity_encoder
        elif self._encoder_name in ['one_hot_encoder']:
            # Fix sequence, then one-hot encode it
            encoder = compose(one_hot_encode, self.__identity_encoder)
        else:
            raise Exception('Encoder name does not exists')
        
        # Prepare if LSTM is used
        if self._architecture_name in ['LSTM']:
            # Encode sequence, then reshape
            self._encoder = compose(self.__lstm_sequence, encoder)
        else:
            self._encoder = encoder
    
    def encode(self, sequence):
        return self._encoder(sequence)
    
    def __fix_sequence_length(self, sequence):
        'Fill a sequence with a dummy value until a fixed length'
        suffix = sequence[:self._memory_length]
        left_len = self._memory_length - len(suffix)
        prefix = [self._num_operators for _ in range(left_len)]
        return prefix + suffix
    
    def __clean_sequence(self, sequence):
        'Keep original values but first -1 element'
        sequence_copy = sequence.copy()
        while len(sequence_copy) > 0 and sequence_copy[0] == -1:
            sequence_copy.pop(0)
        if len(sequence_copy) == 0:
            sequence_copy.append(self._num_operators)  
        return sequence_copy
    
    @staticmethod
    def __lstm_sequence(sequence):
        'Reshape sequence for LSTM architecture usage'
        return [[x] for x in sequence]
    

class ModelPredictorKeras():
    # Keras TensorFlow Artificial Neural Network Model
    def __init__(self, params):
        # Get encoder
        params['memory_length'] = params.get('memory_length', params['num_steps'])
        self._params = params.copy()
        self._encoder = Encoder(params.copy()).encode
        self.__create_keras_model()
    
    def __create_keras_model(self):
        # Create model
        self._model = tf.keras.Sequential()
        self._num_operators = self._params['num_operators']
        architecture_name = self._params['model_architecture']
        input_size = self._params['memory_length']
        hidden_layers = self._params['model_architecture_layers']
        
        # Input layer
        if architecture_name in ['MLP']:
            # MLP input
            self._model.add(tf.keras.Input(shape=input_size))
        elif architecture_name in ['LSTM_Ragged']:
            # Variable length, supported using ragged tensors
            max_length_sequence = self._params['num_steps']
            self._model.add(tf.keras.Input(shape=(max_length_sequence,)))

            # Embedding layer
            if len(hidden_layers) > 0:
                first_layer_size, _, _ = hidden_layers[0]
            else:
                first_layer_size = self._num_operators
            self._model.add(tf.keras.layers.Embedding(self._num_operators + 1, first_layer_size))
        elif architecture_name in ['LSTM']:
            # LSTM input
            self._model.add(tf.keras.Input(shape=(input_size, 1)))

        # Hidden layers
        num_lstm_layers = sum('LSTM' == layer_type for _, _, layer_type in hidden_layers)
        for idx, (layer_size, layer_activation, layer_type) in enumerate(hidden_layers):
            if layer_type == 'Dense':
                self._model.add(tf.keras.layers.Dense(units=layer_size,
                                                      activation=layer_activation))
            elif layer_type == 'LSTM':
                self._model.add(tf.keras.layers.LSTM(units=layer_size,
                                                     activation=layer_activation,
                                                     return_sequences=idx + 1 < num_lstm_layers))        

        # Output layer
        self._model.add(tf.keras.layers.Dense(self._num_operators, activation='softmax'))
    
        # Compile model
        self._model.compile(loss=tf.keras.losses.CategoricalCrossentropy(),
                            optimizer=tf.keras.optimizers.Adam(),
                            metrics=['accuracy'])
    
    def __convert_tensor(self, tensor):
        if self._params['model_architecture'] in ['LSTM_Ragged']:
            return tf.ragged.constant(tensor)
        else:
            return tf.constant(tensor)
        
    def fit(self, X, y, epochs=100, sample_weight=None, 
            verbose=False, early_stopping_params=None, verbose_statistics=False):

        # Pre-process dataset
        X_encoded = [self._encoder(x) for x in X]
        X_tensor = self.__convert_tensor(X_encoded)
        y_tensor = self.__convert_tensor(y)

        # Callbacks
        callbacks = []
        _, _, filename_dict = retrieve_model_info(self._params)

        # History Logger
        if verbose_statistics:
            if not _check_path(filename_dict['model_directory']):
                _create_path(filename_dict['model_directory'])
            history_logger = tf.keras.callbacks.CSVLogger(filename_dict['log_path'],
                                                        separator=',', append=True)
            callbacks.append(history_logger)

        class TimingCallback(tf.keras.callbacks.Callback):
            def __init__(self, logs={}):
                self.logs = []
            def on_epoch_begin(self, epoch, logs={}):
                self.start_time = timer()
            def on_epoch_end(self, epoch, logs={}):
                self.logs.append(timer() - self.start_time)
        timing_cb = TimingCallback()
        if verbose_statistics:
            callbacks.append(timing_cb)
        
        # Early stopping
        if early_stopping_params is not None and verbose_statistics:
            early_stopping = tf.keras.callbacks.EarlyStopping(
                monitor=early_stopping_params['monitor'],
                patience=early_stopping_params['patience'],
                mode=early_stopping_params['mode']
            ) 
            callbacks.append(early_stopping)

        # Train model
        self._model.fit(X_tensor, y_tensor, 
                  epochs=epochs,
                  sample_weight=sample_weight, 
                  verbose=verbose,
                  callbacks=callbacks)
        if verbose_statistics:
            df_times = pd.DataFrame({'time': timing_cb.logs})
            df_times.to_csv(filename_dict['log_time_path'])
            
        # Save predict function
        self._predict = self._model.predict

    def predict(self, sequence):
        # Use model to predict weights
        tensor = self.__convert_tensor([self._encoder(sequence)])
        return self._predict(tensor)[0]
    
    def load(self, model_path=None):
        if model_path is None:
            _, _, filename_dict = retrieve_model_info(self._params)
            model_path = filename_dict['model_path']
            
        if _check_path(model_path):
            self._model = tf.keras.models.load_model(model_path)    
            # Save predict function
            self._predict = self._model.predict
            return True
        else:
            raise Exception(f'model_path "{model_path}" does not exists')

    def save(self, model_path=None):
        if model_path is None:
            _, _, filename_dict = retrieve_model_info(self._params)
            if not _check_path(filename_dict['model_directory']):
                _create_path(filename_dict['model_directory'])
            model_path = filename_dict['model_path']
        self._model.save(model_path)


def ModelPredictor(params):
    # Function that decide which ML model uses.
    # For now, it is only supported the NN architecture with dense 
    # and lstm layers. Future version will support Transformers.
    return ModelPredictorKeras(params)



# -*- coding: utf-8 -*-
"""
This module contains a collection search operators extracted from several well-known metaheuristics from the literature.
All the available operators are listed in ``__all__``

Created on Tue Jan  7 14:54:31 2020

@author: Jorge Mario Cruz-Duarte (jcrvz.github.io), e-mail: jorge.cruz@tec.mx
"""

import numpy as np
import os
from itertools import combinations as _get_combinations
from .population import __selectors__

__all__ = ['local_random_walk', 'random_search', 'random_sample', 'random_flight', 'differential_mutation',
           'firefly_dynamic', 'swarm_dynamic', 'gravitational_search', 'central_force_dynamic', 'spiral_dynamic',
           'genetic_mutation', 'genetic_crossover']


# Search operator aliases
def get_operator_aliases():
    """
    Return two dictionaries with the perturbator and selector aliases for better naming the metaheuristics
    @return: dict, dict
    """
    return {
        'random_search': 'RS',
        'central_force_dynamic': 'CF',
        'differential_mutation': 'DM',
        'firefly_dynamic': 'FD',
        'genetic_crossover': 'GC',
        'genetic_mutation': 'GM',
        'gravitational_search': 'GS',
        'random_flight': 'RF',
        'local_random_walk': 'RW',
        'random_sample': 'RX',
        'spiral_dynamic': 'SD',
        'swarm_dynamic': 'PS'}, {
        'greedy': 'g', 'all': 'd', 'metropolis': 'm', 'probabilistic': 'p'}

# %% SEARCH OPERATORS

def central_force_dynamic(pop, gravity=0.001, alpha=0.01, beta=1.5, dt=1.0):
    """
    Apply the central force dynamic from Central Force Optimisation (CFO) to the population's positions (pop.positions).

    :param population pop: population.
    :param float gravity: optional.
        It is the gravitational constant. The default is 0.001.
    :param float alpha: optional.
        It is the power mass factor. The default is 0.01.
    :param float beta: optional.
        It is the power distance factor. The default is 1.5.
    :param float dt: optional.
        It is the time interval between steps. The default is 1.0.

    :return: None.
    """
    # Initialise acceleration
    acceleration = np.zeros((pop.num_agents, pop.num_dimensions))

    for agent in range(pop.num_agents):
        # Select indices in order to avoid division by zero
        indices = (np.arange(pop.num_agents) != agent)

        # Determine all masses differences with respect to agent
        delta_masses = pop.fitness[indices] - np.tile(
            pop.fitness[agent], (1, pop.num_agents - 1))

        # Determine all vector distances with respect to agent
        delta_positions = pop.positions[indices, :] - np.tile(pop.positions[agent, :], (pop.num_agents - 1, 1))

        distances = np.linalg.norm(delta_positions, 2, 1)

        # Find the quotient part
        quotient = np.heaviside(-delta_masses, 0.0) * (np.abs(delta_masses) ** alpha) / (distances ** beta + 1e-23)

        # Determine the acceleration for each agent
        acceleration[agent, :] = gravity * np.sum(delta_positions * np.tile(
            quotient.transpose(), (1, pop.num_dimensions)), 0)

    pop.positions += 0.5 * acceleration * (dt ** 2)


def differential_crossover(pop, crossover_rate=0.2, version='binomial'):
    """
    Apply the differential crossover from Differential Evolution (DE) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param float crossover_rate: optional.
        Probability factor to perform the crossover. The default is 0.2.
    :param str version: optional.
        Crossover version. It can be 'binomial' or 'exponential'. The default is 'binomial'.

    :return: None.
    """

    # Binomial version
    if version == 'binomial':
        # Define indices
        indices = np.tile(np.arange(pop.num_dimensions), (pop.num_agents, 1))

        # Permute indices per dimension
        rand_indices = np.vectorize(np.random.permutation, signature='(n)->(n)')(indices)

        # Calculate the NOT condition (because positions already updated!)
        condition = np.logical_not(
            (indices == rand_indices) | (np.random.rand(pop.num_agents, pop.num_dimensions) <= crossover_rate))

        # Reverse the ones to their previous positions
        pop.positions[condition] = np.copy(pop.previous_positions[condition])

    # Exponential version
    elif version == 'exponential':
        # Perform the exponential crossover procedure
        for agent in range(pop.num_agents):
            for dim in range(pop.num_dimensions):
                # Initialise L and choose a random index n
                exp_var = 0
                n = np.random.randint(pop.num_dimensions)
                while True:
                    # Increase L and check the exponential CR condition
                    exp_var += 1
                    if np.logical_not((np.random.rand() < crossover_rate) and (exp_var < pop.num_dimensions)):
                        break

                # Perform the crossover if the following condition is met
                if dim not in [(n + x) % pop.num_dimensions for x in range(exp_var)]:
                    pop.positions[agent, dim] = np.copy(pop.previous_positions[agent, dim])

    # Invalid version
    else:
        raise OperatorsError('Invalid differential_crossover version')


def differential_mutation(pop, expression='current-to-best', num_rands=1, factor=1.0):
    """
    Apply the differential mutation from Differential Evolution (DE) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param str expression: optional.
        Type of DE mutation. Available mutations: 'rand', 'best', 'current', 'current-to-best', 'rand-to-best',
         'rand-to-best-and-current'. The default is 'current-to-best'.
    :param int num_rands: optional.
        Number of differences between positions selected at random. The default is 1.
    :param float factor: optional.
        Scale factor (F) to weight contributions from other agents. The default is 1.0.

    :return: None.

    """
    # TODO: Include the expression 'current-to-pbest'
    # Create mutants using the expression provided in scheme
    if expression == 'rand':
        mutant = pop.positions[np.random.permutation(pop.num_agents), :]

    elif expression == 'best':
        mutant = np.tile(pop.global_best_position, (pop.num_agents, 1))

    elif expression == 'current':
        mutant = pop.positions

    elif expression == 'current-to-best':
        mutant = pop.positions + factor * (np.tile(pop.global_best_position, (pop.num_agents, 1)) -
                                           pop.positions[np.random.permutation(pop.num_agents), :])

    elif expression == 'rand-to-best':
        mutant = pop.positions[np.random.permutation(pop.num_agents), :] + factor * (np.tile(
            pop.global_best_position, (pop.num_agents, 1)) - pop.positions[np.random.permutation(pop.num_agents), :])

    elif expression == 'rand-to-best-and-current':
        mutant = pop.positions[np.random.permutation(pop.num_agents), :] + factor * (
                np.tile(pop.global_best_position, (pop.num_agents, 1)) -
                pop.positions[np.random.permutation(pop.num_agents), :] +
                pop.positions[np.random.permutation(pop.num_agents), :] - pop.positions)
    else:
        raise OperatorsError('Invalid DE mutation scheme!')

    # Add random parts according to num_rands
    if num_rands >= 0:
        for _ in range(num_rands):
            mutant += factor * (pop.positions[np.random.permutation(pop.num_agents), :] -
                                pop.positions[np.random.permutation(pop.num_agents), :])
    else:
        raise OperatorsError('Invalid DE mutation scheme!')

    # Replace mutant population in the current one
    pop.positions = np.copy(mutant)


def firefly_dynamic(pop, alpha=1.0, beta=1.0, gamma=100.0, distribution='uniform'):
    """
    Apply the firefly dynamic from Firefly algorithm (FA) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param float alpha: optional.
        Scale of the random value. The default is 1.0.
    :param float beta: optional.
        Scale of the firefly contribution. The default is 1.0.
    :param float gamma: optional.
        Light damping parameters. The default is 100.
    :param str distribution: optional.
        Type of random number. Possible options: 'gaussian', 'uniform', and 'levy'. The default is 'uniform'.

    :return: None.
    """
    # Determine epsilon values
    if distribution == 'gaussian':
        epsilon_value = np.random.standard_normal((pop.num_agents, pop.num_dimensions))

    elif distribution == 'uniform':
        epsilon_value = np.random.uniform(-0.5, 0.5, (pop.num_agents, pop.num_dimensions))
    elif distribution == 'levy':
        epsilon_value = _random_levy((pop.num_agents, pop.num_dimensions), 1.5)
    else:
        raise OperatorsError('Invalid distribution')

    # Initialise delta or difference between two positions
    difference_positions = np.zeros((pop.num_agents, pop.num_dimensions))

    for agent in range(pop.num_agents):
        # Select indices in order to avoid division by zero
        indices = (np.arange(pop.num_agents) != agent)

        # Determine all vector distances with respect to agent
        delta = pop.positions[indices, :] - np.tile(pop.positions[agent, :], (pop.num_agents - 1, 1))

        # Determine differences between lights
        delta_lights = np.tile((pop.fitness[indices] - np.tile(
            pop.fitness[agent], (1, pop.num_agents - 1))).transpose(), (1, pop.num_dimensions))

        # Find the total attraction for each agent
        difference_positions[agent, :] = np.sum(np.heaviside(-delta_lights, 0.0) * delta * np.exp(-gamma * np.tile(
            np.linalg.norm(delta, 2, 1).reshape(pop.num_agents - 1, 1), (1, pop.num_dimensions)) ** 2), 0)

    # Move fireflies according to their attractions
    pop.positions += alpha * epsilon_value + beta * difference_positions


def genetic_crossover(pop, pairing='rank', crossover='blend', mating_pool_factor=0.4):
    """
    Apply the genetic crossover from Genetic Algorithm (GA) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param str pairing: optional.
        It indicates which pairing scheme to employ. Pairing schemes available are: 'cost' (Roulette Wheel or
        Cost Weighting), 'rank' (Rank Weighting), 'tournament', 'random', and 'even-odd'.
        When tournament is chosen, tournament size (tp) and probability (tp) can be encoded such as
        'tournament_{ts}_{tp}', {ts} and {tp}. Writing only 'tournament' is similar to specify 'tournament_3_100'.
        The default is 'rank'.
    :param str crossover: optional.
        It indicates which crossover scheme to employ. Crossover schemes available are: 'single', 'two', 'uniform',
        'blend', and 'linear'. Likewise 'tournament' pairing, coefficients of 'linear' can be enconded such as
        'linear_{coeff1}_{coeff2}' where the offspring is determined as follows:
            ``offspring = coeff1 * father + coeff2 * mother``
        The default is 'blend'.
    :param float mating_pool_factor: optional.
        It indicates the proportion of population to disregard. The default is 0.4.

    :return: None.
    """
    # Mating pool size
    num_mates = int(np.round(mating_pool_factor * pop.num_agents))

    # Get parents (at least a couple per offspring)
    if len(pairing) > 10:  # if pairing = 'tournament_2_100', for example
        pairing, tournament_size, tournament_probability = pairing.split("_")
        tournament_size = int(tournament_size)
        if num_mates < tournament_size:
            num_mates = tournament_size
    else:  # dummy (it must not be used)
        tournament_size, tournament_probability = '3', '100'

    # Number of offsprings (or couples)
    num_couples = pop.num_agents - num_mates

    # Get the mating pool using the natural selection
    mating_pool_indices = np.argsort(pop.fitness)[:num_mates]
    #
    # Roulette Wheel (Cost Weighting) Selection
    if pairing == 'cost':
        # Cost normalisation from mating pool: cost-min(cost @ non mates)
        normalised_cost = pop.fitness[mating_pool_indices] - np.min(
            pop.fitness[np.setdiff1d(np.arange(pop.num_agents), mating_pool_indices)])

        # Determine the related probabilities
        probabilities = np.abs(normalised_cost / (np.sum(normalised_cost) + 1e-23))

        # Perform the roulette wheel selection and return couples
        couple_indices_ = np.searchsorted(np.cumsum(probabilities), np.random.rand(2 * num_couples))

        # Return couples
        couple_indices = couple_indices_.reshape((2, -1))

    # Roulette Wheel (Rank Weighting) Selection
    elif pairing == 'rank':
        # Determine the probabilities
        probabilities = (mating_pool_indices.size - np.arange(
            mating_pool_indices.size)) / np.sum(np.arange(mating_pool_indices.size) + 1)

        # Perform the roulette wheel selection and return couples
        couple_indices_ = np.searchsorted(np.cumsum(probabilities), np.random.rand(2 * num_couples))

        # Return couples
        couple_indices = couple_indices_.reshape((2, -1))

    # Tournament pairing
    elif pairing == 'tournament':
        # Calculate probabilities
        probability = float(tournament_probability) / 100.
        probabilities = probability * ((1 - probability) ** np.arange(tournament_size))

        # Initialise the mother and father indices
        couple_indices = np.full((2, num_couples), np.nan)

        # Perform tournaments until all mates are selected
        for couple in range(num_couples):
            mate = 0
            while mate < 2:
                # Choose tournament candidates
                random_indices = mating_pool_indices[np.random.permutation(mating_pool_indices.size)[:tournament_size]]

                # Determine the candidate fitness values
                candidates_indices = random_indices[np.argsort(pop.fitness[random_indices])]

                # Find the best according to its fitness and probability
                winner = candidates_indices[np.random.rand(tournament_size) < probabilities]
                if winner.size > 0:
                    couple_indices[mate, couple] = int(winner[0])
                    mate += 1

    # Random pairing
    elif pairing == 'random':
        # Return two random indices from mating pool
        couple_indices = mating_pool_indices[np.random.randint(mating_pool_indices.size, size=(2, num_couples))]

    # TODO: Check Even-and-Odd pairing
    # Even-and-Odd pairing
    # elif pairing == "even-odd":
    #     # Check if the num of mates is even
    #     mating_pool_size = mating_pool_indices.size - \
    #         (mating_pool_indices.size % 2)
    #     half_size = mating_pool_size // 2
    #
    #     # Dummy indices according to the mating pool size
    #     remaining = num_couples - half_size
    #     if remaining > 0:
    #         dummy_indices = np.tile(
    #             np.reshape(np.arange(mating_pool_size),
    #                        (-1, 2)).transpose(),
    #             (1, int(np.ceil(num_couples / half_size))))
    #     else:
    #         dummy_indices = np.reshape(np.arange(mating_pool_size),
    #                                    (-1, 2)).transpose()
    #
    #     # Return couple_indices
    #     couple_indices = mating_pool_indices[
    #         dummy_indices[:, :num_couples]]

    # If no pairing procedure recognised
    else:
        raise OperatorsError("Invalid pairing method")

    # Identify offspring indices
    offspring_indices = np.setdiff1d(np.arange(pop.num_agents), mating_pool_indices, True)

    # Prepare crossover variables
    if len(crossover) > 7:  # if crossover = 'linear_0.5_0.5', for example
        cr_split = crossover.split("_")
        if len(cr_split) == 1:
            crossover = cr_split
            coeff1 = coeff2 = 0.5
        elif len(cr_split) == 2:
            crossover = cr_split[0]
            coeff1 = coeff2 = cr_split[1]
        else:
            crossover = cr_split[0]
            coeff1 = cr_split[1]
            coeff2 = cr_split[2]
        coefficients = [float(coeff1), float(coeff2)]
    else:  # dummy (it must not be used)
        coefficients = [np.nan, np.nan]

    # Perform crossover and assign to population
    parent_indices = couple_indices.astype(np.int64)

    # Single-Point Crossover
    if crossover == 'single':
        # Determine the single point per each couple
        single_points = np.tile(np.random.randint(
            pop.num_dimensions, size=parent_indices.shape[1]), (pop.num_dimensions, 1)).transpose()

        # Crossover condition mask
        crossover_mask = np.tile(np.arange(pop.num_dimensions), (parent_indices.shape[1], 1)) <= single_points

        # Get father and mother
        father_position = pop.positions[parent_indices[0, :], :]
        mother_position = pop.positions[parent_indices[1, :], :]

        # Initialise offsprings with mother positions
        offsprings = mother_position
        offsprings[crossover_mask] = father_position[crossover_mask]

    # Two-Point Crossover
    elif crossover == 'two':
        # Find raw points
        raw_points = np.sort(np.random.randint(pop.num_dimensions, size=(parent_indices.shape[1], 2)))

        # Determine the single point per each couple
        points = [np.tile(raw_points[:, x], (pop.num_dimensions, 1)).transpose() for x in range(raw_points.shape[1])]

        # Range matrix
        dummy_matrix = np.tile(np.arange(pop.num_dimensions), (parent_indices.shape[1], 1))

        # Crossover condition mask (only for two points)
        crossover_mask = np.bitwise_or(dummy_matrix <= points[0], dummy_matrix > points[1])

        # Get father and mother
        father_position = pop.positions[parent_indices[0, :], :]
        mother_position = pop.positions[parent_indices[1, :], :]

        # Initialise offsprings with mother positions
        offsprings = mother_position
        offsprings[crossover_mask] = father_position[crossover_mask]

    # Uniform Crossover
    elif crossover == 'uniform':
        # Crossover condition mask (only for uniform crossover)
        crossover_mask = np.random.rand(parent_indices.shape[1], pop.num_dimensions) < 0.5

        # Get father and mother
        father_position = pop.positions[parent_indices[0, :], :]
        mother_position = pop.positions[parent_indices[1, :], :]

        # Initialise offsprings with mother positions
        offsprings = mother_position
        offsprings[crossover_mask] = father_position[crossover_mask]

    # Random blending crossover
    elif crossover == 'blend':
        # Initialise random numbers between 0 and 1
        beta_values = np.random.rand(parent_indices.shape[1], pop.num_dimensions)

        # Get father and mother
        father_position = pop.positions[parent_indices[0, :], :]
        mother_position = pop.positions[parent_indices[1, :], :]

        # Determine offsprings with father and mother positions
        offsprings = beta_values * father_position + (1 - beta_values) * mother_position

    # Linear Crossover: offspring = coeff[0] * father + coeff[1] * mother
    elif crossover == 'linear':
        # Get father and mother
        father_position = pop.positions[parent_indices[0, :], :]
        mother_position = pop.positions[parent_indices[1, :], :]

        # Determine offsprings with father and mother positions
        offsprings = coefficients[0] * father_position + coefficients[1] * mother_position

    # If no crossover method recognised
    else:
        raise OperatorsError('Invalid pairing method')

    # Store offspring positions in the current population
    pop.positions[offspring_indices, :] = np.copy(offsprings)


def genetic_mutation(pop, scale=1.0, elite_rate=0.1, mutation_rate=0.25, distribution='uniform'):
    """
    Apply the genetic mutation from Genetic Algorithm (GA) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param float scale: optional.
        It is the scale factor of the mutations. The default is 1.0.
    :param float elite_rate : optional.
        It is the proportion of population to preserve. The default is 0.1.
    :param float mutation_rate: optional.
        It is the proportion of population to mutate. The default is 0.25.
    :param str distribution: optional.
        It indicates the random distribution that power the mutation. There are only two distribution available:
        'uniform', 'gaussian', and 'levy'. The default is 'uniform'.
    :return: None.
    """

    # Calculate the number of elite agents
    num_elite = int(np.ceil(pop.num_agents * elite_rate))

    # If num_elite equals num_agents then do nothing, or ...
    if num_elite < pop.num_agents:
        # Number of mutations to perform
        num_mutations = int(np.round(pop.num_agents * pop.num_dimensions * mutation_rate))

        # Identify mutable agents
        dimension_indices = np.random.randint(0, pop.num_dimensions, num_mutations)

        if num_elite > 0:
            agent_indices = np.argsort(pop.fitness)[np.random.randint(num_elite, pop.num_agents, num_mutations)]
        else:
            agent_indices = np.random.randint(num_elite, pop.num_agents, num_mutations)

        # Transform indices
        rows, columns = np.meshgrid(agent_indices, dimension_indices)

        # Perform mutation according to the random distribution
        if distribution == 'uniform':
            mutants = np.random.uniform(-1, 1, num_mutations ** 2)

        elif distribution == 'gaussian':
            # Normal with mu = 0 and sigma = parameter
            mutants = np.random.standard_normal(num_mutations ** 2)

        elif distribution == 'levy':
            mutants = _random_levy(num_mutations ** 2, 1.5)

        else:
            raise OperatorsError('Invalid distribution!')

        # Store mutants
        pop.positions[rows.flatten(), columns.flatten()] = scale * mutants


def gravitational_search(pop, gravity=1.0, alpha=0.02):
    """
    Apply the gravitational search from Gravitational Search Algorithm (GSA) to the population's positions
    (pop.positions).

    :param population pop : population.
        It is a population object.
    :param float gravity: optional.
        It is the initial gravitational value. The default is 1.0.
    :param float alpha: optional.
        It is the gravitational damping ratio. The default is 0.02.

    :return: None.
    """

    # Initialise acceleration
    acceleration = np.zeros((pop.num_agents, pop.num_dimensions))

    # Determine the gravitational constant
    gravitation = gravity * np.exp(- alpha * pop.iteration)

    # Determine mass for each agent
    raw_masses = (pop.fitness - np.tile(pop.current_worst_fitness, (1, pop.num_agents)))
    masses = (raw_masses / (np.sum(raw_masses) + 1e-23)).reshape(pop.num_agents)

    for agent in range(pop.num_agents):
        # Select indices in order to avoid division by zero
        indices = (np.arange(pop.num_agents) != agent)

        # Determine all vector distances with respect to agent
        delta_positions = pop.positions[indices, :] - np.tile(pop.positions[agent, :], (pop.num_agents - 1, 1))

        quotient = masses[indices] / (np.linalg.norm(delta_positions, 2, 1) + 1e-23)

        # Force interaction
        force_interaction = gravitation * np.tile(
            quotient.reshape(pop.num_agents - 1, 1), (1, pop.num_dimensions)) * delta_positions

        # Acceleration
        acceleration[agent, :] = np.sum(np.random.rand(pop.num_agents - 1, pop.num_dimensions) * force_interaction, 0)

    # Update velocities
    # TODO: Add different random distributions
    pop.velocities = acceleration + np.random.rand(pop.num_agents, pop.num_dimensions) * pop.velocities

    # Update positions
    pop.positions += pop.velocities


def random_flight(pop, scale=1.0, distribution='levy', beta=1.5):
    """
    Apply the random flight from Random Search (RS) to the population's positions (pop.positions).

    :param population pop : population.
        It is a population object.
    :param float scale: optional.
        It is the step scale. The default is 1.0.
    :param str distribution: optional.
        It is the distribution to draw the random samples. The default is 'levy'.
    :param float beta: optional
        It is the distribution parameter between [1.0, 3.0]. This paramenter only has sense when distribution='levy'.
         The default is 1.5.

    :return: None.
    """

    # Get random samples
    if distribution == 'uniform':
        random_samples = np.random.uniform(
            size=(pop.num_agents, pop.num_dimensions))

    elif distribution == 'gaussian':
        # Normal with mu = 0 and sigma = parameter
        random_samples = np.random.standard_normal(
            (pop.num_agents, pop.num_dimensions))

    elif distribution == 'levy':
        # Calculate the random number with levy stable distribution
        random_samples = _random_levy(size=(pop.num_agents, pop.num_dimensions), beta=beta)

    else:
        raise OperatorsError('Invalid distribution!')

    # Move each agent using levy random displacements
    pop.positions += scale * random_samples * (pop.positions - np.tile(pop.global_best_position, (pop.num_agents, 1)))


def local_random_walk(pop, probability=0.75, scale=1.0, distribution='uniform'):
    """
    Apply the local random walk from Cuckoo Search (CS) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param float probability: optional.
        It is the probability of discovering an alien egg (change an agent's position). The default is 0.75.
    :param float scale: optional.
        It is the step scale. The default is 1.0.
    :param str distribution: optional.
        It is the random distribution used to sample the stochastic variable. The default value is 'uniform'.

    :return: None.
    """

    # Determine random numbers
    if distribution == "uniform":
        r_1 = np.random.rand(pop.num_agents, pop.num_dimensions)
    elif distribution == "gaussian":
        r_1 = np.random.randn(pop.num_agents, pop.num_dimensions)
    elif distribution == "levy":
        r_1 = _random_levy(size=(pop.num_agents, pop.num_dimensions))
    else:
        raise OperatorsError('Invalid distribution!')
    r_2 = np.random.rand(pop.num_agents, pop.num_dimensions)

    # Move positions with a displacement due permutations and probabilities
    pop.positions += scale * r_1 * (pop.positions[
                                    np.random.permutation(pop.num_agents), :] - pop.positions[
                                                                                np.random.permutation(pop.num_agents),
                                                                                :]) * np.heaviside(r_2 - probability,
                                                                                                   0.0)


def random_sample(pop):
    """
    Apply the random_sample to the population's positions (pop.positions). This operator has no memory.

    :param population pop: population.
        It is a population object.

    :return: None.
    """
    # Create random positions using random numbers between -1 and 1
    pop.positions = np.random.uniform(-1, 1, (pop.num_agents, pop.num_dimensions))


def random_search(pop, scale=0.01, distribution='uniform'):
    """
    Apply the random search from Random Search (RS) to the population's positions (pop.positions).

    :param population pop : population.
        It is a population object.
    :param float scale: optional.
        It is the step scale. The default is 0.01.
    :param str distribution: optional.
        It is the distribution used to perform the random search. The default is 'uniform'.

    :return: None.
    """
    # Determine the random step
    if distribution == "uniform":
        random_step = np.random.uniform(-1, 1, (pop.num_agents, pop.num_dimensions))
    elif distribution == "gaussian":
        random_step = np.random.standard_normal((pop.num_agents, pop.num_dimensions))
    elif distribution == "levy":
        random_step = _random_levy(size=(pop.num_agents, pop.num_dimensions))
    else:
        raise OperatorsError('Invalid distribution!')

    # Move each agent using uniform random displacements
    pop.positions += scale * random_step


def spiral_dynamic(pop, radius=0.9, angle=22.5, sigma=0.1):
    """
    Apply the spiral dynamic from Stochastic Spiral Optimisation (SSO) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param float radius: optional.
        It is the convergence rate. The default is 0.9.
    :param float angle: optional.
        Rotation angle (in degrees). The default is 22.5 (degrees).
    :param float sigma: optional.
        Variation of random radii. The default is 0.1.
        Note: if sigma equals 0.0, the operator corresponds to that from the Deterministic Spiral Algorithm.

    :return: None.
    """
    # Determine the rotation matrix
    rotation_matrix = get_rotation_matrix(pop.num_dimensions, np.deg2rad(angle))

    for agent in range(pop.num_agents):
        random_radii = np.random.uniform(radius - sigma, radius + sigma, pop.num_dimensions)
        # If random radii need to be constrained to [0, 1]:
        pop.positions[agent, :] = pop.global_best_position + random_radii * \
                                  np.matmul(rotation_matrix, (pop.positions[agent, :] - pop.global_best_position))


def swarm_dynamic(pop, factor=1.0, self_conf=2.54, swarm_conf=2.56, version='constriction', distribution='uniform'):
    """
    Apply the swarm dynamic from Particle Swarm Optimisation (PSO) to the population's positions (pop.positions).

    :param population pop: population.
        It is a population object.
    :param float factor: optional.
        Inertial or Kappa factor, depending of which PSO version is set. The default is 1.0.
    :param float self_conf: optional.
        Self confidence factor. The default is 2.54.
    :param float swarm_conf: optional.
        Swarm confidence factor. The default is 2.56.
    :param str version: optional.
        Version of the Particle Swarm Optimisation strategy. It can be 'constriction' or 'inertial'. The default is
        'constriction'.
    :param str distribution: optional.
        Distribution to draw the random numbers. It can be 'uniform', 'gaussian', and 'levy'.

    :return: None.
    """
    # Determine random numbers
    if distribution == 'uniform':
        r_1 = np.random.rand(pop.num_agents, pop.num_dimensions)
        r_2 = np.random.rand(pop.num_agents, pop.num_dimensions)
    elif distribution == 'gaussian':
        r_1 = np.random.randn(pop.num_agents, pop.num_dimensions)
        r_2 = np.random.randn(pop.num_agents, pop.num_dimensions)
    elif distribution == 'levy':
        r_1 = _random_levy(size=(pop.num_agents, pop.num_dimensions))
        r_2 = _random_levy(size=(pop.num_agents, pop.num_dimensions))
    else:
        raise OperatorsError('Invalid distribution!')

    # Choose the PSO version = 'inertial' or 'constriction'
    if version == 'inertial':
        # Find new velocities
        pop.velocities = factor * pop.velocities + r_1 * self_conf * (
                pop.particular_best_positions - pop.positions) + \
                         r_2 * swarm_conf * (np.tile(pop.global_best_position, (pop.num_agents, 1)) - pop.positions)
    elif version == 'constriction':
        # Find the constriction factor chi using phi
        phi = self_conf + swarm_conf
        if phi > 4:
            chi = 2 * factor / np.abs(2 - phi - np.sqrt(phi ** 2 - 4 * phi))
        else:
            chi = np.sqrt(factor)

        # Find new velocities
        pop.velocities = chi * (pop.velocities +
                                r_1 * self_conf * (pop.particular_best_positions - pop.positions) +
                                r_2 * swarm_conf * (np.tile(pop.global_best_position, (pop.num_agents, 1)) -
                                                    pop.positions))
    else:
        raise OperatorsError('Invalid swarm_dynamic version')

    # Move each agent using velocity's information
    pop.positions += pop.velocities


# %% INTERNAL METHODS

def _random_levy(size, beta=1.5):
    """
    This is an internal method to draw a random number (or array) using the Levy stable distribution via the
    Mantegna's algorithm.
        R. N. Mantegna and H. E. Stanley, “Stochastic Process with Ultraslow Convergence to a Gaussian: The Truncated
        Levy Flight,” Phys. Rev. Lett., vol. 73, no. 22, pp. 2946–2949, 1994.

    :param size: optional
        Size can be a tuple with all the dimensions. Behaviour similar to ``numpy.random.standard_normal``.
    :param float beta: optional.
        Levy distribution parameter. The default is 1.5.

    :return: numpy.array
    """
    # Calculate x's std dev (Mantegna's algorithm)
    sigma = ((np.math.gamma(1 + beta) * np.sin(np.pi * beta / 2)) / (
            np.math.gamma((1 + beta) / 2) * beta * (2 ** ((beta - 1) / 2)))) ** (1 / beta)

    # Determine x and y using normal distributions with sigma_y = 1
    x = sigma * np.random.standard_normal(size)
    y = np.abs(np.random.standard_normal(size))
    z = np.random.standard_normal(size)

    # Calculate the random number with levy stable distribution
    return z * x / (y ** (1 / beta))


def get_rotation_matrix(dimensions, angle=0.3927):
    """
    Determine the rotation matrix by multiplying all the rotation matrices for each combination of 2D planes.

    :param int dimensions:
        Number of dimensions. Only positive integers greater than one.
    :param float angle: optional.
        Rotation angle (in radians). The default is 0.3927 radians (or 22.5 degrees).

    :return: numpy.array
        Rotation matrix to use over the population positions.
    """
    # Initialise the rotation matrix
    rotation_matrix = np.eye(dimensions)

    # Find the combinations without repetitions
    planes = list(_get_combinations(range(dimensions), 2))

    # Create the rotation matrix
    for xy in range(len(planes)):
        # Read dimensions
        x, y = planes[xy]

        # (Re)-initialise a rotation matrix for each plane
        rotation_plane = np.eye(dimensions)

        # Assign corresponding values
        rotation_plane[[x, x, y, y], [x, y, x, y]] = [np.cos(angle), -np.sin(angle), np.sin(angle), np.cos(angle)]
        rotation_matrix = np.matmul(rotation_matrix, rotation_plane)

    return rotation_matrix


class OperatorsError(Exception):
    """
    Simple OperatorError to manage exceptions.
    """
    pass


# %% TOOLS TO HANDLE THE OPERATORS

def obtain_operators(num_vals=5):
    """
    Generate a list of all the available search operators with a given number of values for each parameter (if so).
    Each element of this list has the following structure:
            search_operator = ('name_of_the_operator',
                               {'parameter1': [value1, value2, ...],
                                'parameter2': [value2, value2, ...],
                                ... },
                               'name_of_the_selector')

    Available selectors from ``population.__selectors__`` are: 'all', 'greedy', 'metropolis', and 'probabilistic'.

    :param int num_vals: optional
        Number of values to generate per each numerical parameter in a search operator. The default is 5.

    :return: list
    """
    return [
        # First line for the initial search operator
        ('random_search', dict(scale=[1.0], distribution=['uniform']), ['greedy']),
        (
            'central_force_dynamic',
            dict(
                gravity=[*np.linspace(0.0, 0.01, num_vals)],
                alpha=[*np.linspace(0.0, 0.01, num_vals)],
                beta=[*np.linspace(1.00, 2.00, num_vals)],
                dt=[*np.linspace(0.0, 2.0, num_vals)]),
            __selectors__),
        (
           'differential_crossover',
           dict(
               crossover_rate=[*np.linspace(0.0, 1.0, num_vals)],
               version=['binomial', 'exponential']),
           __selectors__),
        (
            'differential_mutation',
            dict(
                expression=['rand', 'best', 'current', 'current-to-best', 'rand-to-best', 'rand-to-best-and-current'],
                num_rands=[1, 2, 3],
                factor=[*np.linspace(0.0, 2.5, num_vals)]),
            __selectors__),
        (
            'firefly_dynamic',
            dict(
                distribution=['uniform', 'gaussian', 'levy'],
                alpha=[*np.linspace(0.0, 0.5, num_vals)],
                beta=[*np.linspace(0.01, 1.0, num_vals)],
                gamma=[*np.linspace(1.0, 1000.0, num_vals)]),
            __selectors__),
        (
            'genetic_crossover',
            dict(
                pairing=['rank', 'cost', 'random', 'tournament_2_100', 'tournament_2_75', 'tournament_2_50',
                         'tournament_3_100', 'tournament_3_75', 'tournament_3_50'],
                crossover=['single', 'two', 'uniform', 'blend', 'linear_0.5_0.5'],
                mating_pool_factor=[*np.linspace(0.1, 0.9, num_vals)]),
            __selectors__),
        (
            'genetic_mutation',
            dict(
                scale=[*np.linspace(0.01, 1.0, num_vals)],
                elite_rate=[*np.linspace(0.0, 0.9, num_vals)],
                mutation_rate=[*np.linspace(0.1, 0.9, num_vals)],
                distribution=['uniform', 'gaussian', 'levy']),
            __selectors__),
        (
            'gravitational_search',
            dict(
                gravity=[*np.linspace(0.0, 1.0, num_vals)],
                alpha=[*np.linspace(0.0, 0.04, num_vals)]),
            __selectors__),
        (
            'random_flight',  # Particular case for Levy flight
            dict(
                scale=[*np.linspace(0.01, 1.0, num_vals)],
                distribution=['levy'],
                beta=[*np.linspace(1.00, 2.00, num_vals)]),
            __selectors__),
        (
            'random_flight',
            dict(
                scale=[*np.linspace(0.01, 1.0, num_vals)],
                distribution=['uniform', 'gaussian']),
            __selectors__),
        (
            'local_random_walk',
            dict(
                probability=[*np.linspace(0.01, 0.99, num_vals)],
                scale=[*np.linspace(0.01, 1.0, num_vals)],
                distribution=['uniform', 'gaussian', 'levy']),
            __selectors__),
        (
            'random_sample',
            dict(),
            __selectors__),
        (
            'random_search',
            dict(
                scale=[*np.linspace(0.01, 1.0, num_vals)],
                distribution=['uniform', 'gaussian', 'levy']),
            __selectors__),
        (
            'spiral_dynamic',
            dict(
                radius=[*np.linspace(0.001, 0.999, num_vals)],
                angle=[*np.linspace(0.0, 180.0, num_vals)],
                sigma=[*np.linspace(0.0, 0.5, num_vals)]),
            __selectors__),
        (
            'swarm_dynamic',
            dict(
                factor=[*np.linspace(0.01, 1.0, num_vals)],
                self_conf=[*np.linspace(0.01, 4.99, num_vals)],
                swarm_conf=[*np.linspace(0.01, 4.99, num_vals)],
                version=['inertial', 'constriction'],
                distribution=['uniform', 'gaussian', 'levy']),
            __selectors__)
    ]


def build_operators(heuristics=obtain_operators(), file_name='operators_collection'):
    """
    Create a text file containing a list of all the available search operators, with the same structure as that
    generated by ``operators.obtain_operators``.

    :param list heuristics: optional.
        A list of available search operators. The default is ``obtain_operators()``.
    :param str file_name: optional.
        Customise the file name. The default is 'operators_collection'.

    :return: None.

    """
    # Counters: [classes, methods]
    total_counters = [0, 0]

    # Check if collections exists
    if not os.path.isdir('collections'):
        os.mkdir('collections')

    # Initialise the collection of simple heuristics
    if file_name[-4:] == '.txt':
        file_name = file_name[:-4]
    file = open('collections/' + file_name + '.txt', 'w', encoding='utf-8')

    # For each simple heuristic, read their parameters and values
    for operator, parameters, selectors in heuristics:
        # Update the total classes counter
        total_counters[0] += 1

        # Read the number of parameters and how many values have each one
        num_parameters = len(parameters)
        num_selectors = len(selectors)
        num_combinations: int = 0
        if num_parameters > 0:
            # Read the name and possible values of parameters
            par_names = list(parameters.keys())
            par_values = list(parameters.values())

            # Find the number of values for each parameter
            par_num_values = [np.size(x) for x in par_values]

            # Determine the number of combinations
            num_combinations = int(np.prod(par_num_values))

            # Create the table of all possible combinations (index/parameter)
            indices = [x.flatten() for x in np.meshgrid(*list(map(lambda y: np.arange(y), par_num_values)))]

            # For each combination, create a single dictionary which
            # corresponds to a simple search operator
            for combi in range(num_combinations):
                list_tuples = [
                    (par_names[k], par_values[k][indices[k][combi]])
                    for k in range(num_parameters)]
                simple_par_combination = dict(list_tuples)
                for selector in selectors:
                    file.write("('{}', {}, '{}')\n".format(
                        operator, simple_par_combination, selector))
        elif num_parameters == 0:
            num_combinations = num_selectors
            for selector in selectors:
                file.write("('{}', {}, '{}')\n".format(
                    operator, '{}', selector))

        # Update the total combination counter
        total_counters[1] += num_combinations * num_selectors

        print(f"{operator}: parameters={num_parameters}, " + f"combinations:{num_combinations}")

    # Close the file and print how many types and specific operators were stored
    file.close()
    print("-" * 50 + "--\nTOTAL: families=%d, operators=%d" % tuple(total_counters))


def process_operators(simple_heuristics):
    """
    Decode the list of search operator and deliver two lists, one with the ready-to-execute strings of these operators
    and another with strings of their associated selectors.

    :param list simple_heuristics:
        A list of all the search operators to use. It may look like that saved using ``operators.build_operators``.

    :returns: (list, list)
        out[0] - executable_operators is a list of ready-to-execute string of search operators, and
        out[1] - selectors is the list of strings of the selectors associated to operators.
    """
    # Initialise the list of callable operators (simple heuristics)
    executable_operators = []
    selectors = []

    # For each simple heuristic, read their parameters and values
    for operator, parameters, selector in simple_heuristics:
        # Store selectors
        selectors.append(selector)

        if len(parameters) >= 0:
            sep = ','
            str_parameters = []

            for parameter, value in parameters.items():

                # Check if a value is string
                if type(value) == str:
                    str_parameters.append("{}='{}'".format(parameter, value))
                else:
                    str_parameters.append('{}={}'.format(parameter, value))

            # Create an executable string with given arguments
            full_string = '{}({})'.format(operator, sep.join(str_parameters))
        else:
            # Create an executable string with default arguments
            full_string = '{}()'.format(operator)

        # Store the read operator
        executable_operators.append(full_string)

    # Return two lists of executable operators and selectors
    return executable_operators, selectors


# %% AUTOMATIC RUN

if __name__ == '__main__':
    """
    Automatically create a collection of search operators using ```operators.obtain_operators(num_vals=5)``` and
    save it as 'automatic.txt'.
    """
    build_operators(obtain_operators(num_vals=5), file_name='automatic')



from .experiment import Experiment

def test_brute_force():
  exp = Experiment(f"./exconf/tests/brute_force.json")
  exp.run()

def test_basic_mhs():
  exp = Experiment(f"./exconf/tests/basic_mhs.json")
  exp.run()

def test_static():
  exp = Experiment(f"./exconf/tests/static.json")
  exp.run()

def test_dynamic():
  exp = Experiment(f"./exconf/tests/dynamic.json")
  exp.run()

def test_neural_network():
  exp = Experiment(f"./exconf/tests/neural_network.json")
  exp.run()
    


# -*- coding: utf-8 -*-
import setuptools
import pathlib

# The text of the README file
README = (pathlib.Path(__file__).parent / 'README.md').read_text()

setuptools.setup(
    name='customhys',
    version='1.1.4',
    packages=setuptools.find_packages(),
    url='https://github.com/jcrvz/customhys',
    license='MIT License',
    author='Jorge Mario Cruz-Duarte',
    author_email='jorge.cruz@tec.mx',
    description='This framework provides tools for solving, but not limited to, continuous optimisation problems '
                'using a hyper-heuristic approach for customising metaheuristics. ',
    long_description_content_type='text/markdown',
    long_description=README,
    keywords=['metaheuristics', 'hyper-heuristic', 'optimization', 'automatic design', 'global optimization',
              'evolutionary computation', 'bio-inspired', 'algorithm design'],
    python_requires=">=3.8",
    install_requires=[
        "numpy>=1.22.0",
        "scipy>=1.5.0",
        "matplotlib>=3.2.2",
        "tqdm>=4.47.0",
        "scikit-learn>=1.2.2",
        "pandas>=1.5.3",
    ],
    extras_require={
        "ML": [
            "tensorflow>=2.8.0; sys_platform != 'darwin'",
            "tensorflow-macos>=2.11.0; sys_platform == 'darwin'",
            "tensorflow-metal>=0.7.1; sys_platform == 'darwin'",
        ]
    },
    include_package_data=True,

)
